@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2004-2011 uim Project http://code.google.com/p/uim/
+  Copyright (c) 2004-2012 uim Project http://code.google.com/p/uim/
 
   All rights reserved.
 
@@ -495,7 +495,7 @@ word_window_add(WordWindow *window)
     g_free(phonetic);
     g_free(literal);
     g_free(cclass_desc);
-    g_free(word);
+    free(word);
 #endif
   } else {
     ret = 0;
