@@ -1,5 +1,5 @@
 ;;;
-;;; Copyright (c) 2003-2008 uim Project http://code.google.com/p/uim/
+;;; Copyright (c) 2003-2012 uim Project http://code.google.com/p/uim/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -58,6 +58,29 @@
 (define skk-ddskk-like-heading-label-char-list '("a" "s" "d" "f" "j" "k" "l"))
 (define skk-uim-heading-label-char-list '("1" "2" "3" "4" "5" "6" "7" "8" "9" "0"))
 
+(define skk-ja-rk-rule (append ja-rk-rule-basic ja-rk-rule-additional))
+(define skk-okuri-char-alist '())
+(define skk-downcase-alist '())
+(define skk-set-henkan-point-key '())
+
+(define skk-ichar-downcase
+  (lambda (x)
+    (or (cdr (or (assoc x skk-downcase-alist)
+                 '(#f . #f)))
+        (ichar-downcase x))))
+
+(define skk-ichar-upper-case?
+  (lambda (x)
+    (or (if (assoc x skk-downcase-alist) #t #f) (ichar-upper-case? x))))
+
+(define skk-context-set-okuri-head-using-alist!
+  (lambda (sc s)
+    (skk-context-set-okuri-head!
+      sc
+      (or (cdr (or (assoc s skk-okuri-char-alist)
+                   '(#f . #f)))
+          s))))
+
 ;; style specification
 (define skk-style-spec
   '(;; (style-element-name . validator)
@@ -124,6 +147,8 @@
 
 (define skk-input-rule-roma 0)
 (define skk-input-rule-azik 1)
+(define skk-input-rule-act 2)
+(define skk-input-rule-kzik 3)
 
 (define skk-child-type-editor 0)
 (define skk-child-type-dialog 1)
@@ -147,7 +172,8 @@
 (define skk-show-cursor-on-preedit? #f)
 (define skk-show-candidates-with-okuri? #f)
 
-(define skk-dic-init #f)
+(define skk-dic #f)
+(define skk-context-list '())
 
 (define skk-prepare-activation
   (lambda (sc)
@@ -265,6 +291,37 @@
 		     (skk-prepare-activation dsc)
 		     (skk-set-rule! dsc skk-input-rule-azik))))
 
+(register-action 'action_skk_act
+		 (lambda (sc)
+		   '(ja_act
+		     ""
+		     "ACT"
+		     "ACTĥ޻ϥ⡼"))
+		 (lambda (sc)
+		   (let ((dsc (skk-find-descendant-context sc)))
+		     (= (skk-context-input-rule dsc)
+			skk-input-rule-act)))
+		 (lambda (sc)
+		   (let ((dsc (skk-find-descendant-context sc)))
+		     (skk-prepare-activation dsc)
+		     (skk-set-rule! dsc skk-input-rule-act))))
+
+
+(register-action 'action_skk_kzik
+		 (lambda (sc)
+		   '(ja_kzik
+		     ""
+		     "KZIK"
+		     "KZIKĥ޻ϥ⡼"))
+		 (lambda (sc)
+		   (let ((dsc (skk-find-descendant-context sc)))
+		     (= (skk-context-input-rule dsc)
+			skk-input-rule-kzik)))
+		 (lambda (sc)
+		   (let ((dsc (skk-find-descendant-context sc)))
+		     (skk-prepare-activation dsc)
+		     (skk-set-rule! dsc skk-input-rule-kzik))))
+
 ;; Update widget definitions based on action configurations. The
 ;; procedure is needed for on-the-fly reconfiguration involving the
 ;; custom API
@@ -313,10 +370,28 @@
     (let ((rkc (skk-context-rk-context sc))
 	  (rule (cond
 		 ((= input-rule skk-input-rule-roma)
-		  ja-rk-rule)
+                  (set! skk-okuri-char-alist '())
+                  (set! skk-downcase-alist '()) 
+                  (set! skk-set-henkan-point-key '())
+		  skk-ja-rk-rule)
 		 ((= input-rule skk-input-rule-azik)
 		  (require "japanese-azik.scm")
-		  ja-azik-rule))))
+                  (set! skk-okuri-char-alist ja-azik-skk-okuri-char-alist)
+                  (set! skk-downcase-alist ja-azik-skk-downcase-alist) 
+                  (set! skk-set-henkan-point-key ja-azik-skk-set-henkan-point-key)
+		  ja-azik-rule)
+		 ((= input-rule skk-input-rule-act)
+		  (require "japanese-act.scm")
+                  (set! skk-okuri-char-alist ja-act-skk-okuri-char-alist)
+                  (set! skk-downcase-alist ja-act-skk-downcase-alist)
+                  (set! skk-set-henkan-point-key ja-act-skk-set-henkan-point-key)
+		  ja-act-rule)
+		 ((= input-rule skk-input-rule-kzik)
+		  (require "japanese-kzik.scm")
+                  (set! skk-okuri-char-alist '())
+                  (set! skk-downcase-alist '())
+                  (set! skk-set-henkan-point-key '())
+		  ja-kzik-rule))))
       (skk-context-set-input-rule! sc input-rule)
       (rk-context-set-rule! rkc rule))))
 
@@ -337,13 +412,16 @@
 (define skk-read-personal-dictionary
   (lambda ()
     (if (not (setugid?))
-	(or (skk-lib-read-personal-dictionary skk-uim-personal-dic-filename)
-	    (skk-lib-read-personal-dictionary skk-personal-dic-filename)))))
+        (or (skk-lib-read-personal-dictionary skk-dic
+                                              skk-uim-personal-dic-filename)
+            (skk-lib-read-personal-dictionary skk-dic
+                                              skk-personal-dic-filename)))))
 
 (define skk-save-personal-dictionary
   (lambda ()
     (if (not (setugid?))
-	(skk-lib-save-personal-dictionary skk-uim-personal-dic-filename))))
+        (skk-lib-save-personal-dictionary skk-dic
+                                          skk-uim-personal-dic-filename))))
 
 (define skk-flush
   (lambda (sc)
@@ -369,22 +447,23 @@
 
 (define skk-context-new
   (lambda (id im)
-    (if (not skk-dic-init)
+    (if (not skk-dic)
 	(let ((hostname (if skk-skkserv-use-env?
 			    (or (getenv "SKKSERVER") "localhost")
 			    skk-skkserv-hostname)))
-	  (set! skk-dic-init #t)
 	  (if skk-use-recursive-learning?
 	      (require "skk-editor.scm"))
 	  (require "skk-dialog.scm")
-	  (skk-lib-dic-open
-	   skk-dic-file-name skk-use-skkserv? hostname skk-skkserv-portnum
-	   skk-skkserv-address-family)
+          (set! skk-dic (skk-lib-dic-open skk-dic-file-name
+                                          skk-use-skkserv?
+                                          hostname
+                                          skk-skkserv-portnum
+                                          skk-skkserv-address-family))
           (if skk-use-look?
               (skk-lib-look-open skk-look-dict))
 	  (skk-read-personal-dictionary)))
     (let ((sc (skk-context-new-internal id im))
-	  (rkc (rk-context-new ja-rk-rule #t #f)))
+	  (rkc (rk-context-new skk-ja-rk-rule #t #f)))
       (skk-context-set-widgets! sc skk-widgets)
       (skk-context-set-head! sc '())
       (skk-context-set-rk-context! sc rkc)
@@ -502,11 +581,12 @@
   (lambda (sc n)
     (let* ((head (skk-context-head sc))
     	   (cand (skk-lib-get-nth-candidate
-		  n
-		  (skk-make-string head skk-type-hiragana)
-		  (skk-context-okuri-head sc)
-		  (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
-		  skk-use-numeric-conversion?)))
+                  skk-dic
+                  n
+                  (cons (skk-make-string head skk-type-hiragana)
+                        (skk-context-okuri-head sc))
+                  (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
+                  skk-use-numeric-conversion?)))
       (if skk-show-annotation?
 	  cand
 	  (skk-lib-remove-annotation cand)))))
@@ -520,6 +600,7 @@
 (define skk-get-nth-completion
   (lambda (sc n)
     (skk-lib-get-nth-completion
+     skk-dic
      n
      (skk-make-string (skk-context-head sc) skk-type-hiragana)
      skk-use-numeric-conversion?
@@ -579,8 +660,9 @@
 	   (res (string-append cand okuri appendix))
 	   (head (skk-context-head sc)))
       (skk-lib-commit-candidate
-       (skk-make-string head skk-type-hiragana)
-       (skk-context-okuri-head sc)
+       skk-dic
+       (cons (skk-make-string head skk-type-hiragana)
+             (skk-context-okuri-head sc))
        (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
        (skk-context-nth sc)
        skk-use-numeric-conversion?)
@@ -593,11 +675,13 @@
 (define skk-purge-candidate
   (lambda (sc)
     (let ((res (skk-lib-purge-candidate
-		 (skk-make-string (skk-context-head sc) skk-type-hiragana)
-		 (skk-context-okuri-head sc)
-		 (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
-		 (skk-context-nth sc)
-		 skk-use-numeric-conversion?)))
+                skk-dic
+                (cons
+                 (skk-make-string (skk-context-head sc) skk-type-hiragana)
+                 (skk-context-okuri-head sc))
+                (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
+                (skk-context-nth sc)
+                skk-use-numeric-conversion?)))
       (if res
 	  (skk-save-personal-dictionary))
       (skk-reset-candidate-window sc)
@@ -624,10 +708,11 @@
 	 (skk-context-set-dcomp-word!
 	  sc
 	  (skk-lib-get-dcomp-word
-	   (skk-make-string
-	    (skk-context-head sc) (skk-context-kana-mode sc))
-	   skk-use-numeric-conversion?
-	   skk-use-look?))))))
+           skk-dic
+           (skk-make-string
+            (skk-context-head sc) (skk-context-kana-mode sc))
+           skk-use-numeric-conversion?
+           skk-use-look?))))))
 
 (define skk-append-okuri-string
   (lambda (sc str)
@@ -653,11 +738,12 @@
 (define skk-begin-conversion
   (lambda (sc)
     (let ((res (skk-lib-get-entry
-	       (skk-make-string (skk-context-head sc) skk-type-hiragana)
-		(skk-context-okuri-head sc)
-		(skk-make-string (skk-context-okuri sc)
-				 skk-type-hiragana)
-		skk-use-numeric-conversion?)))
+                skk-dic
+                (skk-make-string (skk-context-head sc) skk-type-hiragana)
+                (skk-context-okuri-head sc)
+                (skk-make-string (skk-context-okuri sc)
+                                 skk-type-hiragana)
+                skk-use-numeric-conversion?)))
       (if res
 	  (begin
 	    (skk-context-set-nth! sc 0)
@@ -676,6 +762,7 @@
     (if (eq? (skk-context-state sc) 'skk-state-kanji)
 	(skk-append-residual-kana sc))
     (skk-lib-get-completion
+     skk-dic
      (skk-make-string (skk-context-head sc) (skk-context-kana-mode sc))
      skk-use-numeric-conversion?
      skk-use-look?)
@@ -992,7 +1079,7 @@
 (define skk-proc-state-direct
   (lambda (c key key-state)
     (let* ((sc (skk-find-descendant-context c))
-	   (key-str (charcode->string (ichar-downcase key)))
+	   (key-str (charcode->string (skk-ichar-downcase key)))
 	   (rkc (skk-context-rk-context sc))
 	   (res #f)
 	   (kana (skk-context-kana-mode sc)))
@@ -1031,21 +1118,21 @@
        ;; 1. commits "n" as kana according to kana-mode
        ;; 2. switch mode by "{L,l,/,Q,C-q,C-Q,q}"
        (if (and (skk-wide-latin-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (begin
 	     (set! res (rk-push-key-last! rkc))
 	     (skk-context-set-state! sc 'skk-state-wide-latin)
 	     #f)
 	   #t)
        (if (and (skk-latin-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (begin
 	     (set! res (rk-push-key-last! rkc))
 	     (skk-context-set-state! sc 'skk-state-latin)
 	     #f)
 	   #t)
        (if (and (skk-latin-conv-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (let* ((residual-kana (rk-push-key-last! rkc)))
 	     (if residual-kana
 		 (skk-commit sc (skk-get-string sc residual-kana kana)))
@@ -1054,7 +1141,7 @@
 	     #f)
 	   #t)
        (if (and (skk-kanji-mode-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (let* ((residual-kana (rk-push-key-last! rkc)))
 	     (if residual-kana
 		 (skk-commit sc (skk-get-string sc residual-kana kana)))
@@ -1063,7 +1150,7 @@
 	     #f)
 	   #t)
        (if (and (skk-hankaku-kana-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (let* ((kana (skk-context-kana-mode sc))
 		  (new-kana (if (= kana skk-type-hankana)
 		    		  skk-type-hiragana
@@ -1073,7 +1160,7 @@
 	     #f)
 	   #t)
        (if (and (skk-kana-toggle-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (begin
 	     (set! res (rk-push-key-last! rkc))
 	     (skk-context-kana-toggle sc)
@@ -1085,7 +1172,7 @@
        ;; 2. commits " " as native space (such as Qt::Key_Space)
        ;;    unless expected rkc list includes " "
        (if (and (skk-plain-space-key? key key-state)
-		(not (string-find (rk-expect rkc) key-str)))
+		(not (rk-expect-key? rkc key-str)))
 	   (begin
 	     (set! res (rk-push-key-last! rkc))
 	     (skk-commit-raw-with-preedit-update sc key key-state)
@@ -1107,35 +1194,25 @@
 	     (skk-commit-raw-with-preedit-update sc key key-state)
 	     #f)
 	   #t)
-       (if (ichar-upper-case? key)
+       (if (skk-ichar-upper-case? key)
 	   (if (and 
 		(skk-rk-pending? sc)
 		(not (rk-current-seq rkc)))
 	       ;; ddskk compatible behavior but not in SKK speciation
-	       (begin
+	       (let ((str (rk-push-key! rkc (charcode->string
+					      (skk-ichar-downcase key)))))
 		 (skk-context-set-state! sc 'skk-state-kanji)
-		 (skk-append-string sc (rk-push-key!
-					rkc
-					(charcode->string
-					 (ichar-downcase key))))
+		 (if str
+		   (skk-append-string sc str))
 		 #f)
 	       (let* ((residual-kana (rk-push-key-last! rkc)))
 		 ;; handle preceding "n"
 		 (if residual-kana
 		     (skk-commit sc (skk-get-string sc residual-kana kana)))
 		 (skk-context-set-state! sc 'skk-state-kanji)
-		 (set! key (ichar-downcase key))
+		 (set! key (skk-ichar-downcase key))
 		 #t))
 	   #t)
-       ;; Hack to handle "n1" sequence as "1".
-       ;; This should be handled in rk.scm. -- ekato
-       (if (and (not (ichar-alphabetic? key))
-		(not (string-find (rk-expect rkc) key-str)))
-	   (let* ((residual-kana (rk-push-key-last! rkc)))
-	     (if residual-kana
-		 (skk-commit sc (skk-get-string sc residual-kana kana)))
-	     #t)
-	   #t)
        ;; bad strategy. see bug #528
        (if (symbol? key)
 	   (begin
@@ -1197,7 +1274,8 @@
 	   (not (skk-rk-pending? sc))
 	   (not (string=? (skk-context-dcomp-word sc) "")))
       (if (skk-lib-get-entry
-	   (skk-context-dcomp-word sc) "" "" skk-use-numeric-conversion?)
+           skk-dic
+           (skk-context-dcomp-word sc) "" "" skk-use-numeric-conversion?)
 	  (begin
 	    (skk-string-list-to-context-head
 	     sc
@@ -1213,22 +1291,24 @@
       (skk-append-residual-kana sc)
       (if (not (null? (skk-context-head sc)))
 	  (let ((dcomp (skk-lib-get-dcomp-word
-			(skk-make-string
-			 (skk-context-head sc)
-			 (skk-context-kana-mode sc))
-			skk-use-numeric-conversion?
-			skk-use-look?)))
+                        skk-dic
+                        (skk-make-string
+                         (skk-context-head sc)
+                         (skk-context-kana-mode sc))
+                        skk-use-numeric-conversion?
+                        skk-use-look?)))
 	    (if (not (string=? dcomp ""))
 		(begin
 		  (skk-string-list-to-context-head
 		   sc
 		   (string-to-list dcomp))
 		  (if (skk-lib-get-entry
-		       (skk-make-string
-			(skk-context-head sc) skk-type-hiragana)
-		       ""
-		       ""
-		       skk-use-numeric-conversion?)
+                       skk-dic
+                       (skk-make-string
+                        (skk-context-head sc) skk-type-hiragana)
+                       ""
+                       ""
+                       skk-use-numeric-conversion?)
 		      (begin
 			(skk-context-set-nth! sc 0)
 			(skk-commit sc (skk-prepare-commit-string sc)))
@@ -1237,11 +1317,12 @@
 			(skk-flush sc))))
 		(begin
 		  (if (skk-lib-get-entry
-		       (skk-make-string
-			(skk-context-head sc) skk-type-hiragana)
-		       ""
-		       ""
-		       skk-use-numeric-conversion?)
+                       skk-dic
+                       (skk-make-string
+                        (skk-context-head sc) skk-type-hiragana)
+                       ""
+                       ""
+                       skk-use-numeric-conversion?)
 		      (begin
 			(skk-context-set-nth! sc 0)
 			(skk-commit sc (skk-prepare-commit-string sc)))
@@ -1255,11 +1336,12 @@
       (if (not (null? (skk-context-head sc)))
 	  (begin
 	    (if (skk-lib-get-entry
-		 (skk-make-string
-		  (skk-context-head sc) skk-type-hiragana)
-		 ""
-		 ""
-		 skk-use-numeric-conversion?)
+                 skk-dic
+                 (skk-make-string
+                  (skk-context-head sc) skk-type-hiragana)
+                 ""
+                 ""
+                 skk-use-numeric-conversion?)
 		(begin
 		  (skk-context-set-nth! sc 0)
 		  (skk-commit sc (skk-prepare-commit-string sc)))
@@ -1311,12 +1393,13 @@
 		  sc
 		  (if (not (skk-rk-pending? sc))
 		      (skk-lib-get-dcomp-word
-		       (skk-make-string
-			(skk-context-head sc)
-			(skk-context-kana-mode sc))
-		       skk-use-numeric-conversion?
-		       skk-use-look?)
-			"")))
+                       skk-dic
+                       (skk-make-string
+                        (skk-context-head sc)
+                        (skk-context-kana-mode sc))
+                       skk-use-numeric-conversion?
+                       skk-use-look?)
+                      "")))
 	     #f)
 	   #t)
        (if (or
@@ -1353,11 +1436,12 @@
 	       (skk-append-residual-kana sc)
 	       (let ((sl (string-to-list
 			  (skk-lib-get-dcomp-word
-			   (skk-make-string
-			    (skk-context-head sc)
-			    (skk-context-kana-mode sc))
-			   skk-use-numeric-conversion?
-			   skk-use-look?))))
+                           skk-dic
+                           (skk-make-string
+                            (skk-context-head sc)
+                            (skk-context-kana-mode sc))
+                           skk-use-numeric-conversion?
+                           skk-use-look?))))
 		 (if (not (null? sl))
 		     (begin
 		       (skk-string-list-to-context-head sc sl)
@@ -1426,18 +1510,18 @@
 	     (skk-begin-conversion sc)
 	     #f)
 	   #t)
-       (if (and (ichar-upper-case? key)
+       (if (and (skk-ichar-upper-case? key)
 		(not (null? (skk-context-head sc))))
-	   (let ((key-str (charcode->string (ichar-downcase key))))
+	   (let ((key-str (charcode->string (skk-ichar-downcase key))))
 	     (set! res (skk-rk-push-key-match-without-new-seq rkc key-str))
 	     (if (and
 		  (skk-rk-pending? sc)
 		  (not (rk-current-seq rkc))
 		  res)
 		 ;; ddskk compatible behavior but not in SKK speciation
-		 (begin	
+		 (begin
 		   (skk-context-set-state! sc 'skk-state-okuri)
-		   (skk-context-set-okuri-head!
+		   (skk-context-set-okuri-head-using-alist!
 		    sc
 		    (car (reverse (rk-context-seq rkc))))
 		   (rk-context-set-seq! rkc '())
@@ -1446,9 +1530,9 @@
 		   #f)
 		 (begin
 		   (skk-context-set-state! sc 'skk-state-okuri)
-		   (set! key (ichar-downcase key))
-		   (skk-context-set-okuri-head! sc key-str)
-		   (if (skk-sokuon-shiin-char? key)
+		   (set! key (skk-ichar-downcase key))
+		   (skk-context-set-okuri-head-using-alist! sc key-str)
+		   (if (and (not (member key skk-set-henkan-point-key)) (skk-sokuon-shiin-char? key))
 		       (begin
 			 (set! res (rk-push-key! rkc key-str))
 			 (if res
@@ -1481,22 +1565,8 @@
 		   (skk-flush sc)))
 	     #f)
 	   #t)
-       ;; Hack to handle "n1" sequence as "1".
-       ;; This should be handled in rk.scm. -- ekato
-       (if (and (not (ichar-alphabetic? key))
-		(not (string-find
-		      (rk-expect rkc)
-		      (charcode->string (ichar-downcase key)))))
-	   (let* ((residual-kana (rk-push-key-last! rkc)))
-	     (if residual-kana
-		 (skk-context-set-head! sc
-					(cons
-					 residual-kana
-					 (skk-context-head sc))))
-	     #t)
-	   #t)
        (begin
-	 (set! key (ichar-downcase key))  
+	 (set! key (skk-ichar-downcase key))  
 	 (set! stat (skk-context-state sc))
 	 (set! res
 	       (rk-push-key!
@@ -1556,10 +1626,11 @@
 	  (skk-context-set-nr-candidates!
 	   sc
 	   (skk-lib-get-nr-candidates
-	    (skk-make-string (skk-context-head sc) skk-type-hiragana)
-	    (skk-context-okuri-head sc)
-	    (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
-	    skk-use-numeric-conversion?))
+            skk-dic
+            (skk-make-string (skk-context-head sc) skk-type-hiragana)
+            (skk-context-okuri-head sc)
+            (skk-make-string (skk-context-okuri sc) skk-type-hiragana)
+            skk-use-numeric-conversion?))
 	  (im-activate-candidate-selector
 	   sc
 	   (cond
@@ -1750,9 +1821,10 @@
     (if incr
 	(begin
 	  (if (> (- (skk-lib-get-nr-completions
-		     (skk-make-string (skk-context-head sc) skk-type-hiragana)
-		     skk-use-numeric-conversion?
-		     skk-use-look?)
+                     skk-dic
+                     (skk-make-string (skk-context-head sc) skk-type-hiragana)
+                     skk-use-numeric-conversion?
+                     skk-use-look?)
 		    1)
 		 (skk-context-completion-nth sc))
 	      (skk-context-set-completion-nth!
@@ -1815,14 +1887,15 @@
 		  (sl (string-to-list comp)))
 	     (if (not (null? sl))
 		 (begin (skk-lib-get-completion
-			 (skk-get-current-completion sc)
-			 skk-use-numeric-conversion?
-			 skk-use-look?)))
+                         skk-dic
+                         (skk-get-current-completion sc)
+                         skk-use-numeric-conversion?
+                         skk-use-look?)))
 	     (skk-lib-clear-completions
 	      (skk-make-string
 	       (skk-context-head sc)
 	       skk-type-hiragana)
-	       skk-use-numeric-conversion?)
+	      skk-use-numeric-conversion?)
 	     (if (not (null? sl))
 		 (skk-string-list-to-context-head sc sl))
 	     (skk-context-set-completion-nth! sc 0)
@@ -1982,7 +2055,7 @@
 	 (set! res
 	       (rk-push-key!
 		rkc
-		(charcode->string (ichar-downcase key))))
+		(charcode->string (skk-ichar-downcase key))))
 	 (if (and res
 	 	  (or
 		   (list? (car res))
@@ -1993,7 +2066,7 @@
 		   (skk-begin-conversion sc)))
 	     (begin
 	       (if (= (length (rk-context-seq rkc)) 1)
-		   (skk-context-set-okuri-head! sc (charcode->string key)))))))
+		   (skk-context-set-okuri-head-using-alist! sc (charcode->string key)))))))
       #f)))
 
 (define skk-proc-state-latin
@@ -2057,11 +2130,18 @@
   (lambda (id im arg)
     (let ((sc (skk-context-new id im)))
       (update-style skk-style-spec (symbol-value skk-style))
+      (set! skk-context-list (cons sc skk-context-list))
       sc)))
 
 (define skk-release-handler
   (lambda (sc)
-    (skk-save-personal-dictionary)))
+    (skk-save-personal-dictionary)
+    (set! skk-context-list (delete! sc skk-context-list))
+    (if (null? skk-context-list)
+      (begin
+        (skk-lib-look-close)
+        (skk-lib-free-dic skk-dic)
+        (set! skk-dic #f)))))
 
 (define skk-press-key-handler
   (lambda (sc key state)
